import pyvisa
from time import sleep
# import pkg_resources
import os
from typing import Callable

from ..base import VoltageMeasurementDevice
from ...utility.visa import enumerate_devices

class Keithley2700(VoltageMeasurementDevice):
    """
    Wrapper class for the Keithley2700 SMU controlled via pyvisa
    """
    
    def __init__(self, instr, check_front_switch=True):
        self.instr = instr
        if check_front_switch:
            self._check_front_input_selected()
        self.buffer_next_idx = None
                
    def __del__(self):
        """Properly close the instrument connection"""
        self.instr.close()
        
    def _check_front_input_selected(self):
        """
        Make sure the front switch selecting the inputs selects the FRONT inputs
        Raises
        ------
        Exception
            If front input state != 1

        Returns
        -------
        switch : str
            Front input state
        """
        switch = self.instr.query("SYSTem:FRSwitch?").strip("\n")
        if switch != "1":
            raise Exception("The Keithley's INPUT switch must select the [F]ront inputs")
        return switch
    
        
    def query(self, query):
        try:
            return self.instr.query(query).strip("\n")
        except pyvisa.VisaIOError as e:
            print(f"VisaIOError raised during query: '{query}'")
            raise e
    
    def query_int(self, query):
        return int(float(self.query(query)))
        
    # RUN COMMANDS ON THE DEVICE
    def run(self, code, verbose=False):
        """
        Run SCPI code on the device by writing it.
        Empty lines, leading whitespaces and lines starting with ' or # are ignored.
        Parameters
        ----------
        code : str
            SCPI commands
        """
        script = ''
        for line in code.strip(" ").split("\n"):
            l = line.strip(" ")
            if len(l) == 0 or l[0] in "#'": continue
            script += l + "\n"
        if verbose:
            print(f"Running code:\n{script}")
        try:
            self.instr.write(script)
        except pyvisa.VisaIOError as e:
            print(f"VisaIOError raised while writing command(s):\n'{script}'\n")
            raise e
    
    def run_script(self, script_path, verbose=False):       
        """
        Load the code from script_path and run it via self.run
        
        Parameters
        ----------
        script_path : str
            full path to the script
        verbose : bool, optional
            If true, print a message when the script is run. The default is False.

        Returns
        -------
        None.
        """
        with open(os.path.expanduser(script_path), "r") as file:
            script = file.read()
        
        if verbose: print(f"Running script: {script_path}")
        self.run(script)
    
    def beep(self, length=0.5, pitch=1000):
        """The Keithley2700 cant beep on command :("""
        pass


    def reset(self, verbose=False):
        """
        Reset the device
        """
        
        reset_script = """
        VOLT:DC:RANGe:AUTO ON
        ' DC Voltage measurement
        SENSe:FUNC 'VOLT:DC'
        ' Set voltage divider if required
        ' SENSE:VOLT:DC:IDIVider OFF
        ' Disable continuous initiation
        INIT:CONT OFF
        ' Disable Buffer and trigger
        TRACe:FEED NONE 
        TRACe:FEED:CONTrol NEVer
        TRIGger:SOURce IMMediate
        ' Set timestamp format to relative
        SYSTem:TSTamp:TYPE RELative
        """
        self.run(reset_script)
        # self.run_script(scripts["instrument_reset"], verbose=verbose)
        self.buffer_clear()

    
    # INTERACT WITH DEVICE BUFFER
    # might not be needed
    def buffer_clear(self):
        buffer_reset_script = """
        TRACe:CLEar
        TRACe:CLEar:AUTO ON
        """
        self.run(buffer_reset_script)
        
    def buffer_get_size(self):
        return self.query_int("TRACe:POINts?")
    
    def buffer_set_size(self, s):
        if not type(s) == int or s < 2 or s > 55000:
            raise ValueError(f"Invalid buffer size: {s}. Must be int and between 2 and 55000")
        self.instr.write(f"TRACe:POINts {s}")
        
    def buffer_measure(self, interval: float=0.5, verbose=False):
        if interval < 0.001 or interval > 999999.999:
            raise ValueError("Interval must be between 0.001 and 999999.999")
        self.run(f"""
        TRIGger:TIMer {interval}
        TRIGger:COUNt INFinity
        TRIGger:SOURce TIMer
        """)
        self.buffer_clear()
        self.run("""
        TRACe:FEED SENSe
        ' write continuously
        TRACe:FEED:CONTrol ALWays
        INITiate:CONTinuous ON
        """)
        self.buffer_next_idx = 0
        self.buffer_size = self.buffer_get_size()
        if verbose: print("Started buffer measurement")
    
    def buffer_read_new_values(self):
        if self.buffer_next_idx is None:
            raise ValueError("You must first start a buffer measurement by calling buffer_measure()")
        # TRACe:NEXT? returns the next index that will be written to
        new_next_idx = self.query_int("TRACe:NEXT?")
        if new_next_idx == self.buffer_next_idx:
            raise ValueError(f"No new value or buffer has been filled completely (next reading at {self.buffer_next_idx})")
        # if reached end of buffer, first read to end and then from beginning
        vals = ""
        if new_next_idx < self.buffer_next_idx:
            count = self.buffer_size - self.buffer_next_idx
            # print(f"start={self.buffer_next_idx}, stop={new_next_idx}, count={count}")
            vals += self.query(f"TRACe:DATA:SELected? {self.buffer_next_idx}, {count}").strip("\n")
            self.buffer_next_idx = 0
        count = new_next_idx - self.buffer_next_idx
        if count > 0:
            # print(f"start={self.buffer_next_idx}, stop={new_next_idx}, count={count}")
            vals += self.query(f"TRACe:DATA:SELected? {self.buffer_next_idx}, {count}").strip("\n")
        self.buffer_next_idx = new_next_idx
        processed_vals = [self.process_reading(val) for val in vals.strip("#").split("#")]
        return processed_vals
        
    
    # MEASUREMENT
    @staticmethod
    def process_reading(reading: str):
        """
        process a reading. Only works with VDC and relative time stamps right now!
        '-1.19655066E+01VDC,+9627.275SECS,+64993RDNG\n'
        May have trailing commas and trailing '#' characters
        
        Returns
        -------
        [timestamp, voltage]
        """
        parts = reading.strip("#").strip(",").split(",")
        if len(parts) != 3:
            raise ValueError(f"Invalid reading: '{reading}'")
        vdc = float(parts[0][:-3])
        timestamp = float(parts[1][:-4])
        # TODO last value
        return timestamp, vdc        
        
    def read_value(self):
        """
        Read a single value

        Returns
        -------
        [timestamp, voltage]
        """
        reading = self.instr.query("READ?")
        return self.process_reading(reading)
    
    def measure(self, interval: int, update_func: Callable[None, [int, float, float]]|None=None, max_measurements:int|None=None):
        """
        Take voltage readings after <interval> milliseconds.

        Parameters
        ----------
        interval : int
            Number of milliseconds to wait between readings.
        update_func : Callable[None, [int, float, float]] or None, optional
            A function that is called after each reading with parameters <n_reading>, <time>, <voltage>. The default is None.
        max_measurements : int or None, optional
           Number of readings to perform. Set to None for infinite. The default is None.

        Returns
        -------
        None.

        """
        self.reset(verbose=True)
        try:
            i = 0
            while max_measurements is None or i < max_measurements:
                tval, vval = self.read_value()
                if update_func:
                    update_func(i, tval, vval)
                sleep(interval)
                i += 1
        except KeyboardInterrupt:
            pass
        print("Measurement stopped" + " "*50)

    @staticmethod
    def enumerate_devices(query="(GPIB)|(USB)?*:INSTR"):
        return enumerate_devices("MODEL 2700", query)

    @staticmethod
    def connect_device(name):
        rm = pyvisa.ResourceManager()
        instr = rm.open_resource(name)
        return Keithley2700(instr)

    def __str__(self):
        return "Keithley 2700"
